﻿#define COMPOSITE_DEVICE    // 複合デバイスの場合は定義する

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Win32.SafeHandles;
using System.Runtime.InteropServices;

namespace USB_IR_Library
{
    public class USBIR
    {

        //Constant definitions from setupapi.h, which we aren't allowed to include directly since this is C#
        internal const uint DIGCF_PRESENT = 0x02;
        internal const uint DIGCF_DEVICEINTERFACE = 0x10;
        ////Constants for CreateFile() and other file I/O functions
        //internal const short FILE_ATTRIBUTE_NORMAL = 0x80;
        //internal const short INVALID_HANDLE_VALUE = -1;
        internal const uint GENERIC_READ = 0x80000000;
        internal const uint GENERIC_WRITE = 0x40000000;
        //internal const uint CREATE_NEW = 1;
        //internal const uint CREATE_ALWAYS = 2;
        internal const uint OPEN_EXISTING = 3;
        internal const uint FILE_SHARE_READ = 0x00000001;
        internal const uint FILE_SHARE_WRITE = 0x00000002;
        ////Constant definitions for certain WM_DEVICECHANGE messages
        //internal const uint WM_DEVICECHANGE = 0x0219;
        //internal const uint DBT_DEVICEARRIVAL = 0x8000;
        //internal const uint DBT_DEVICEREMOVEPENDING = 0x8003;
        //internal const uint DBT_DEVICEREMOVECOMPLETE = 0x8004;
        //internal const uint DBT_CONFIGCHANGED = 0x0018;
        ////Other constant definitions
        internal const uint DBT_DEVTYP_DEVICEINTERFACE = 0x05;
        internal const uint DEVICE_NOTIFY_WINDOW_HANDLE = 0x00;
        internal const uint ERROR_SUCCESS = 0x00;
        internal const uint ERROR_NO_MORE_ITEMS = 0x00000103;
        internal const uint SPDRP_HARDWAREID = 0x00000001;

        public const uint IR_FREQ_MIN = 25000;                  // 赤外線周波数設定最小値 25KHz
        public const uint IR_FREQ_MAX = 50000;                  // 赤外線周波数設定最大値 50KHz
        public const uint IR_SEND_DATA_USB_SEND_MAX_LEN = 14;   // USB送信１回で送信する最大ビット数
        public const uint IR_SEND_DATA_MAX_LEN = 300;           // 赤外線送信データ設定最大長[byte]

        //public const int FORMAT_AEHA = 0x00000001;          // 家電フォーマット
        //public const int FORMAT_NEC = 0x00000002;           // NECフォーマット
        //public const int FORMAT_SONY = 0x00000003;          // SONYフォーマット
        //public const int FORMAT_MITSUBISHI = 0x00000004;    // MITSUBISHIフォーマット
        public const uint FORMAT_AEHA_READER_CODE = 0x007B003D;         // 家電協 Reader Code ON:3.2ms/26us=123(0x7B), OFF:1.6ms/26us=61(0x3D)
        public const uint FORMAT_AEHA_BIT_OFF = 0x000F000F;             // 家電協 BitOFF ON:0.4ms/26us=15(0x0F), OFF:0.4ms/26us=15(0x0F)
        public const uint FORMAT_AEHA_BIT_ON = 0x000F002E;              // 家電協 BitON  ON:0.4ms/26us=15(0x0F), OFF:1.2ms/26us=46(0x2E)
        public const uint FORMAT_NEC_READER_CODE = 0x015A00AD;          // NEC Reader Code ON:9.0ms/26us=346(0x15A), OFF:4.5ms/26us=173(0xAD)
        public const uint FORMAT_NEC_BIT_OFF = 0x00160016;              // NEC BitOFF ON:0.56ms/26us=22(0x16), OFF:0.56ms/26us=22(0x16)
        public const uint FORMAT_NEC_BIT_ON = 0x00160041;               // NEC BitON  ON:0.56ms/26us=22(0x16), OFF:1.68ms/26us=65(0x41)
        public const uint FORMAT_SONY_READER_CODE = 0x005C0017;         // SONY Reader Code ON:2.4ms/26us=92(0x5C), OFF:0.6ms/26us=23(0x17)
        public const uint FORMAT_SONY_BIT_OFF = 0x00170017;             // SONY BitOFF ON:0.6ms/26us=23(0x17), OFF:0.6ms/26us=23(0x17)
        public const uint FORMAT_SONY_BIT_ON = 0x002E0017;              // SONY BitON  ON:1.2ms/26us=46(0x2E), OFF:0.6ms/26us=23(0x17)
        public const uint FORMAT_MITSU_READER_CODE = 0x00000000;        // MITSUBISHI Reader Code なし
        public const uint FORMAT_MITSU_BIT_OFF = 0x000F001F;            // MITSUBISHI BitOFF ON:0.4ms/26us=15(0x0F), OFF:0.8ms/26us=31(0x1F)
        public const uint FORMAT_MITSU_BIT_ON = 0x000F004D;             // MITSUBISHI BitON  ON:0.4ms/26us=15(0x0F), OFF:2.0ms/26us=77(0x4D)
        public const uint FORMAT_STOP_CODE = 0x00170619;                // STOP CODE
        
        public enum IR_FORMAT
        {
            AEHA = 1,
            NEC,
            SONY,
            MITSUBISHI
        }
        public enum PLARAIL_BAND
        {
            BAND_A = 1,
            BAND_B
        }
        public enum PLARAIL_DIRECTION
        {
            FORWARD = 1,
            BACKWARD
        }

        //Various structure definitions for structures that this code will be using
        internal struct SP_DEVICE_INTERFACE_DATA
        {
            internal uint cbSize;               //DWORD
            internal Guid InterfaceClassGuid;   //GUID
            internal uint Flags;                //DWORD
            internal uint Reserved;             //ULONG_PTR MSDN says ULONG_PTR is "typedef unsigned __int3264 ULONG_PTR;"  
        }

        internal struct SP_DEVICE_INTERFACE_DETAIL_DATA
        {
            internal uint cbSize;               //DWORD
            internal char[] DevicePath;         //TCHAR array of any size
        }

        internal struct SP_DEVINFO_DATA
        {
            internal uint cbSize;       //DWORD
            internal Guid ClassGuid;    //GUID
            internal uint DevInst;      //DWORD
            internal uint Reserved;     //ULONG_PTR  MSDN says ULONG_PTR is "typedef unsigned __int3264 ULONG_PTR;"  
        }

        internal struct DEV_BROADCAST_DEVICEINTERFACE
        {
            internal uint dbcc_size;            //DWORD
            internal uint dbcc_devicetype;      //DWORD
            internal uint dbcc_reserved;        //DWORD
            internal Guid dbcc_classguid;       //GUID
            internal char[] dbcc_name;          //TCHAR array
        }

        //DLL Imports.  Need these to access various C style unmanaged functions contained in their respective DLL files.
        //--------------------------------------------------------------------------------------------------------------
        //Returns a HDEVINFO type for a device information set.  We will need the 
        //HDEVINFO as in input parameter for calling many of the other SetupDixxx() functions.
        [DllImport("setupapi.dll", SetLastError = true, CharSet = CharSet.Unicode)]
        internal static extern IntPtr SetupDiGetClassDevs(
            ref Guid ClassGuid,     //LPGUID    Input: Need to supply the class GUID. 
            IntPtr Enumerator,      //PCTSTR    Input: Use NULL here, not important for our purposes
            IntPtr hwndParent,      //HWND      Input: Use NULL here, not important for our purposes
            uint Flags);            //DWORD     Input: Flags describing what kind of filtering to use.

        //Gives us "PSP_DEVICE_INTERFACE_DATA" which contains the Interface specific GUID (different
        //from class GUID).  We need the interface GUID to get the device path.
        [DllImport("setupapi.dll", SetLastError = true, CharSet = CharSet.Unicode)]
        internal static extern bool SetupDiEnumDeviceInterfaces(
            IntPtr DeviceInfoSet,           //Input: Give it the HDEVINFO we got from SetupDiGetClassDevs()
            IntPtr DeviceInfoData,          //Input (optional)
            ref Guid InterfaceClassGuid,    //Input 
            uint MemberIndex,               //Input: "Index" of the device you are interested in getting the path for.
            ref SP_DEVICE_INTERFACE_DATA DeviceInterfaceData);    //Output: This function fills in an "SP_DEVICE_INTERFACE_DATA" structure.

        //SetupDiDestroyDeviceInfoList() frees up memory by destroying a DeviceInfoList
        [DllImport("setupapi.dll", SetLastError = true, CharSet = CharSet.Unicode)]
        internal static extern bool SetupDiDestroyDeviceInfoList(
            IntPtr DeviceInfoSet);          //Input: Give it a handle to a device info list to deallocate from RAM.

        //SetupDiEnumDeviceInfo() fills in an "SP_DEVINFO_DATA" structure, which we need for SetupDiGetDeviceRegistryProperty()
        [DllImport("setupapi.dll", SetLastError = true, CharSet = CharSet.Unicode)]
        internal static extern bool SetupDiEnumDeviceInfo(
            IntPtr DeviceInfoSet,
            uint MemberIndex,
            ref SP_DEVINFO_DATA DeviceInterfaceData);

        //SetupDiGetDeviceRegistryProperty() gives us the hardware ID, which we use to check to see if it has matching VID/PID
        [DllImport("setupapi.dll", SetLastError = true, CharSet = CharSet.Unicode)]
        internal static extern bool SetupDiGetDeviceRegistryProperty(
            IntPtr DeviceInfoSet,
            ref SP_DEVINFO_DATA DeviceInfoData,
            uint Property,
            ref uint PropertyRegDataType,
            IntPtr PropertyBuffer,
            uint PropertyBufferSize,
            ref uint RequiredSize);

        //SetupDiGetDeviceInterfaceDetail() gives us a device path, which is needed before CreateFile() can be used.
        [DllImport("setupapi.dll", SetLastError = true, CharSet = CharSet.Unicode)]
        internal static extern bool SetupDiGetDeviceInterfaceDetail(
            IntPtr DeviceInfoSet,                   //Input: Wants HDEVINFO which can be obtained from SetupDiGetClassDevs()
            ref SP_DEVICE_INTERFACE_DATA DeviceInterfaceData,                    //Input: Pointer to an structure which defines the device interface.  
            IntPtr DeviceInterfaceDetailData,      //Output: Pointer to a SP_DEVICE_INTERFACE_DETAIL_DATA structure, which will receive the device path.
            uint DeviceInterfaceDetailDataSize,     //Input: Number of bytes to retrieve.
            ref uint RequiredSize,                  //Output (optional): The number of bytes needed to hold the entire struct 
            IntPtr DeviceInfoData);                 //Output (optional): Pointer to a SP_DEVINFO_DATA structure

        //Overload for SetupDiGetDeviceInterfaceDetail().  Need this one since we can't pass NULL pointers directly in C#.
        [DllImport("setupapi.dll", SetLastError = true, CharSet = CharSet.Unicode)]
        internal static extern bool SetupDiGetDeviceInterfaceDetail(
            IntPtr DeviceInfoSet,                   //Input: Wants HDEVINFO which can be obtained from SetupDiGetClassDevs()
            ref SP_DEVICE_INTERFACE_DATA DeviceInterfaceData,               //Input: Pointer to an structure which defines the device interface.  
            IntPtr DeviceInterfaceDetailData,       //Output: Pointer to a SP_DEVICE_INTERFACE_DETAIL_DATA structure, which will contain the device path.
            uint DeviceInterfaceDetailDataSize,     //Input: Number of bytes to retrieve.
            IntPtr RequiredSize,                    //Output (optional): Pointer to a DWORD to tell you the number of bytes needed to hold the entire struct 
            IntPtr DeviceInfoData);                 //Output (optional): Pointer to a SP_DEVINFO_DATA structure

        //Need this function for receiving all of the WM_DEVICECHANGE messages.  See MSDN documentation for
        //description of what this function does/how to use it. Note: name is remapped "RegisterDeviceNotificationUM" to
        //avoid possible build error conflicts.
        [DllImport("user32.dll", SetLastError = true, CharSet = CharSet.Unicode)]
        internal static extern IntPtr RegisterDeviceNotification(
            IntPtr hRecipient,
            IntPtr NotificationFilter,
            uint Flags);

        //Takes in a device path and opens a handle to the device.
        [DllImport("kernel32.dll", SetLastError = true, CharSet = CharSet.Unicode)]
        static extern SafeFileHandle CreateFile(
            string lpFileName,
            uint dwDesiredAccess,
            uint dwShareMode,
            IntPtr lpSecurityAttributes,
            uint dwCreationDisposition,
            uint dwFlagsAndAttributes,
            IntPtr hTemplateFile);

        //Uses a handle (created with CreateFile()), and lets us write USB data to the device.
        [DllImport("kernel32.dll", SetLastError = true, CharSet = CharSet.Unicode)]
        static extern bool WriteFile(
            SafeFileHandle hFile,
            byte[] lpBuffer,
            uint nNumberOfBytesToWrite,
            ref uint lpNumberOfBytesWritten,
            IntPtr lpOverlapped);

        //Uses a handle (created with CreateFile()), and lets us read USB data from the device.
        [DllImport("kernel32.dll", SetLastError = true, CharSet = CharSet.Unicode)]
        static extern bool ReadFile(
            SafeFileHandle hFile,
            IntPtr lpBuffer,
            uint nNumberOfBytesToRead,
            ref uint lpNumberOfBytesRead,
            IntPtr lpOverlapped);

        ////--------------- Global Varibles Section ------------------
        ////USB related variables that need to have wide scope.
        static String DevicePath = null;   //Need the find the proper device path before you can open file handles.

        //Globally Unique Identifier (GUID) for HID class devices.  Windows uses GUIDs to identify things.
        static Guid InterfaceClassGuid = new Guid(0x4d1e55b2, 0xf16f, 0x11cf, 0x88, 0xcb, 0x00, 0x11, 0x11, 0x00, 0x00, 0x30);
        //--------------- End of Global Varibles ------------------


        public static SafeFileHandle openUSBIR(IntPtr hRecipient)
        {
            SafeFileHandle HandleToUSBDevice = null;
            try
            {
                //Register for WM_DEVICECHANGE notifications.  This code uses these messages to detect plug and play connection/disconnection events for USB devices
                DEV_BROADCAST_DEVICEINTERFACE DeviceBroadcastHeader = new DEV_BROADCAST_DEVICEINTERFACE();
                DeviceBroadcastHeader.dbcc_devicetype = DBT_DEVTYP_DEVICEINTERFACE;
                DeviceBroadcastHeader.dbcc_size = (uint)Marshal.SizeOf(DeviceBroadcastHeader);
                DeviceBroadcastHeader.dbcc_reserved = 0;	//Reserved says not to use...
                DeviceBroadcastHeader.dbcc_classguid = InterfaceClassGuid;

                //Need to get the address of the DeviceBroadcastHeader to call RegisterDeviceNotification(), but
                //can't use "&DeviceBroadcastHeader".  Instead, using a roundabout means to get the address by 
                //making a duplicate copy using Marshal.StructureToPtr().
                IntPtr pDeviceBroadcastHeader = IntPtr.Zero;  //Make a pointer.
                pDeviceBroadcastHeader = Marshal.AllocHGlobal(Marshal.SizeOf(DeviceBroadcastHeader)); //allocate memory for a new DEV_BROADCAST_DEVICEINTERFACE structure, and return the address 
                Marshal.StructureToPtr(DeviceBroadcastHeader, pDeviceBroadcastHeader, false);  //Copies the DeviceBroadcastHeader structure into the memory already allocated at DeviceBroadcastHeaderWithPointer
                RegisterDeviceNotification(hRecipient, pDeviceBroadcastHeader, DEVICE_NOTIFY_WINDOW_HANDLE);
                //RegisterDeviceNotification(this.Handle, pDeviceBroadcastHeader, DEVICE_NOTIFY_WINDOW_HANDLE);

                //Now make an initial attempt to find the USB device, if it was already connected to the PC and enumerated prior to launching the application.
                //If it is connected and present, we should open read and write handles to the device so we can communicate with it later.
                //If it was not connected, we will have to wait until the user plugs the device in, and the WM_DEVICECHANGE callback function can process
                //the message and again search for the device.
                if (CheckIfPresentAndGetUSBDevicePath())	//Check and make sure at least one device with matching VID/PID is attached
                {
                    uint ErrorStatusWrite;

                    //We now have the proper device path, and we can finally open read and write handles to the device.
                    HandleToUSBDevice = CreateFile(DevicePath, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, IntPtr.Zero, OPEN_EXISTING, 0, IntPtr.Zero);
                    ErrorStatusWrite = (uint)Marshal.GetLastWin32Error();

                    if (ErrorStatusWrite == ERROR_SUCCESS)
                    {
                    }
                    else //for some reason the device was physically plugged in, but one or both of the read/write handles didn't open successfully...
                    {

                        HandleToUSBDevice = null;

                    }
                }
                else	//Device must not be connected (or not programmed with correct firmware)
                {
                }
            }
            catch
            {
            }

            return HandleToUSBDevice;
        }
        public static int closeUSBIR(SafeFileHandle HandleToUSBDevice)
        {
            int i_ret = -1;
            try
            {
                if (HandleToUSBDevice != null)
                {
                    HandleToUSBDevice.Close();
                    i_ret = 0;
                }
            }
            catch
            {
            }
            return i_ret;
        }
        public static int writeUSBIR(SafeFileHandle HandleToUSBDevice, IR_FORMAT format_type, byte[] code, int code_len)
        {
            int i_ret = -1;

            try
            {
                int code_len_check = (int)(code_len / 8);
                if ((code_len % 8) > 0)
                {
                    code_len_check++;
                }
                // パラメータチェック
                if (HandleToUSBDevice != null && 0 < code_len && code_len <= (IR_SEND_DATA_MAX_LEN*8) && code.Length >= code_len_check)
                {
                    uint reader_code = 0;
                    uint bit_0 = 0;
                    uint bit_1 = 0;
                    uint stop_code = FORMAT_STOP_CODE;

                    switch(format_type)
                    {
                        case IR_FORMAT.AEHA:
                            reader_code = FORMAT_AEHA_READER_CODE;
                            bit_0 = FORMAT_AEHA_BIT_OFF;
                            bit_1 = FORMAT_AEHA_BIT_ON;
                            break;
                        case IR_FORMAT.NEC:
                            reader_code = FORMAT_NEC_READER_CODE;
                            bit_0 = FORMAT_NEC_BIT_OFF;
                            bit_1 = FORMAT_NEC_BIT_ON;
                            break;
                        case IR_FORMAT.SONY:
                            reader_code = FORMAT_SONY_READER_CODE;
                            bit_0 = FORMAT_SONY_BIT_OFF;
                            bit_1 = FORMAT_SONY_BIT_ON;
                            break;
                        case IR_FORMAT.MITSUBISHI:
                            reader_code = FORMAT_MITSU_READER_CODE;
                            bit_0 = FORMAT_MITSU_BIT_OFF;
                            bit_1 = FORMAT_MITSU_BIT_ON;
                            break;
                        default:
                            break;
                    }
                    if (writeUSBIRCode(HandleToUSBDevice, 38000, reader_code, bit_0, bit_1, stop_code, code, (uint)code_len) == 0)
                    {
                        i_ret = 0;
                    }
                    else
                    {
                        i_ret = -3;
                    }
                }
                else
                {   // パラメータエラー
                    i_ret = -2;
                }
            }
            catch
            {
                i_ret = 1;
            }
            return i_ret;
        }
        public static int writeUSBIRCode(SafeFileHandle HandleToUSBDevice, uint freq, uint reader_code, uint bit_0, uint bit_1, uint stop_code, byte[] code, uint bit_len)
        {
            int i_ret = -1;
            byte[] outbuffer = new byte[65];
            byte[] inbuffer = new byte[65];
            uint BytesWritten = 0;
            uint BytesRead = 0;
            bool error_flag = false;
            uint send_bit_num = 1 + bit_len + 1;    // 送信ビット数　リーダーコード(1) + コード + 終了コード(1)
            uint send_bit_pos = 0;                  // 送信セット済みビット位置
            uint set_bit_size = 0;

            try
            {
                int code_len_check = (int)(bit_len / 8);
                if((bit_len % 8) > 0)
                {
                    code_len_check++;
                }
                // パラメータチェック
                if (HandleToUSBDevice != null && IR_FREQ_MIN <= freq && freq <= IR_FREQ_MAX && 0 < bit_len && bit_len <= (IR_SEND_DATA_MAX_LEN*8) && code.Length >= code_len_check)
                {
                    // データセット
                    while (true)
                    {
                        outbuffer[0] = 0x00;
                        outbuffer[1] = 0x34;
                        //送信総ビット数
                        outbuffer[2] = (byte)((send_bit_num >> 8) & 0xFF);
                        outbuffer[3] = (byte)(send_bit_num & 0xFF);
                        outbuffer[4] = (byte)((send_bit_pos >> 8) & 0xFF);
                        outbuffer[5] = (byte)(send_bit_pos & 0xFF);
                        if (send_bit_num > send_bit_pos)
                        {
                            set_bit_size = send_bit_num - send_bit_pos;
                            if (set_bit_size > IR_SEND_DATA_USB_SEND_MAX_LEN)
                            {
                                set_bit_size = IR_SEND_DATA_USB_SEND_MAX_LEN;
                            }
                        }
                        else
                        {   // 送信データなし
                            set_bit_size = 0;
                        }
                        outbuffer[6] = (byte)(set_bit_size & 0xFF);

                        if (set_bit_size > 0)
                        {
                            // データセット
                            // 赤外線コードコピー
                            for (uint fi = 0; fi < set_bit_size; fi++)
                            {
                                if (send_bit_pos == 0)
                                {   // Reader Code
                                    // ON Count
                                    outbuffer[7 + (fi * 4)] = (byte)((reader_code >> 24) & 0xFF);
                                    outbuffer[7 + (fi * 4) + 1] = (byte)((reader_code >> 16) & 0xFF);
                                    // OFF Count
                                    outbuffer[7 + (fi * 4) + 2] = (byte)((reader_code >> 8) & 0xFF);
                                    outbuffer[7 + (fi * 4) + 3] = (byte)(reader_code & 0xFF);
                                }
                                else if (send_bit_pos == (send_bit_num - 1))
                                {   // Stop Code
                                    // ON Count
                                    outbuffer[7 + (fi * 4)] = (byte)((stop_code >> 24) & 0xFF);
                                    outbuffer[7 + (fi * 4) + 1] = (byte)((stop_code >> 16) & 0xFF);
                                    // OFF Count
                                    outbuffer[7 + (fi * 4) + 2] = (byte)((stop_code >> 8) & 0xFF);
                                    outbuffer[7 + (fi * 4) + 3] = (byte)(stop_code & 0xFF);
                                }
                                else
                                {
                                    int byte_pos = (int)(send_bit_pos - 1) / 8;
                                    int bit_pos = (int)(send_bit_pos - 1) % 8;
                                    int bit_mask = 0x01 << bit_pos;
                                    int tmp_data = code[byte_pos];

                                    if ((tmp_data & bit_mask & 0xFF) != 0)
                                    {   // Bit 1
                                        // ON Count
                                        outbuffer[7 + (fi * 4)] = (byte)((bit_1 >> 24) & 0xFF);
                                        outbuffer[7 + (fi * 4) + 1] = (byte)((bit_1 >> 16) & 0xFF);
                                        // OFF Count
                                        outbuffer[7 + (fi * 4) + 2] = (byte)((bit_1 >> 8) & 0xFF);
                                        outbuffer[7 + (fi * 4) + 3] = (byte)(bit_1 & 0xFF);
                                    }
                                    else
                                    {   // bit 0
                                        // ON Count
                                        outbuffer[7 + (fi * 4)] = (byte)((bit_0 >> 24) & 0xFF);
                                        outbuffer[7 + (fi * 4) + 1] = (byte)((bit_0 >> 16) & 0xFF);
                                        // OFF Count
                                        outbuffer[7 + (fi * 4) + 2] = (byte)((bit_0 >> 8) & 0xFF);
                                        outbuffer[7 + (fi * 4) + 3] = (byte)(bit_0 & 0xFF);
                                    }
                                }
                                send_bit_pos++;
                            }
                            if (WriteFile(HandleToUSBDevice, outbuffer, 65, ref BytesWritten, IntPtr.Zero))
                            {
                                //Now get the response packet from the firmware.
                                inbuffer[0] = 0;
                                {
                                    if (ReadFileManagedBuffer(HandleToUSBDevice, inbuffer, 65, ref BytesRead, IntPtr.Zero))	//Blocking function, unless an "overlapped" structure is used
                                    {
                                        //INBuffer[0] is the report ID, which we don't care about.
                                        //INBuffer[1] is an echo back of the command (see microcontroller firmware).
                                        //INBuffer[2] contains the I/O port pin value for the pushbutton (see microcontroller firmware).  
                                        if (inbuffer[1] == 0x34)
                                        {
                                            if (inbuffer[2] == 0x00)
                                            {   // OK
                                            }
                                            else
                                            {
                                                // NG
                                                error_flag = true;
                                            }
                                        }
                                    }
                                    else
                                    {
                                        // NG
                                        error_flag = true;
                                    }
                                }
                            }
                            else
                            {
                                // NG
                                error_flag = true;
                            }

                        }
                        else
                        {   // 送信データなし
                            break;
                        }
                    }

                    // データ送信要求セット
                    if (error_flag == false)
                    {
                        outbuffer[0] = 0;			//The first byte is the "Report ID" and does not get sent over the USB bus.  Always set = 0.
                        outbuffer[1] = 0x35;		//0x81 is the "Get Pushbutton State" command in the firmware
                        outbuffer[2] = (byte)((freq >> 8) & 0xFF);
                        outbuffer[3] = (byte)(freq & 0xFF);
                        outbuffer[4] = (byte)((send_bit_num >> 8) & 0xFF);
                        outbuffer[5] = (byte)(send_bit_num & 0xFF);

                        //To get the pushbutton state, first, we send a packet with our "Get Pushbutton State" command in it.
                        if (WriteFile(HandleToUSBDevice, outbuffer, 65, ref BytesWritten, IntPtr.Zero))	//Blocking function, unless an "overlapped" structure is used
                        {
                            //Now get the response packet from the firmware.
                            inbuffer[0] = 0;
                            {
                                if (ReadFileManagedBuffer(HandleToUSBDevice, inbuffer, 65, ref BytesRead, IntPtr.Zero))	//Blocking function, unless an "overlapped" structure is used
                                {
                                    //INBuffer[0] is the report ID, which we don't care about.
                                    //INBuffer[1] is an echo back of the command (see microcontroller firmware).
                                    //INBuffer[2] contains the I/O port pin value for the pushbutton (see microcontroller firmware).  
                                    if (inbuffer[1] == 0x35)
                                    {
                                        if (inbuffer[2] == 0x00)
                                        {   // OK
                                            i_ret = 0;
                                        }
                                        else
                                        {
                                            // NG
                                            error_flag = true;
                                        }
                                    }
                                }
                                else
                                {
                                    // NG
                                    i_ret = -5;
                                }
                            }
                        }
                        else
                        {
                            // NG
                            i_ret = -4;
                        }
                    }
                    else
                    {   // データセットエラー
                        i_ret = -3;
                    }
                }
                else
                {   // パラメータエラー
                    i_ret = -2;
                }
            }
            catch
            {
                i_ret = 1;
            }
            return i_ret;
        }
        public static int writeUSBIRCode(SafeFileHandle HandleToUSBDevice, uint freq, uint reader_code, uint bit_0, uint bit_1, uint stop_code, byte[] code, uint bit_len, uint[] repeat_code, uint repeat_code_len, uint repeat_code_send_num)
        {
            int i_ret = -1;
            byte[] outbuffer = new byte[65];
            byte[] inbuffer = new byte[65];
            uint BytesWritten = 0;
            uint BytesRead = 0;
            bool error_flag = false;
            uint send_bit_num = 1 + bit_len + 1 + (repeat_code_len * repeat_code_send_num);    // 送信ビット数　リーダーコード(1) + コード + 終了コード(1) + (リピートコードビット*繰り返し回数)
            uint send_bit_pos = 0;                  // 送信セット済みビット位置
            uint set_bit_size = 0;

            try
            {
                int code_len_check = (int)(bit_len / 8);
                if ((bit_len % 8) > 0)
                {
                    code_len_check++;
                }
                // パラメータチェック
                if (HandleToUSBDevice != null && IR_FREQ_MIN <= freq && freq <= IR_FREQ_MAX && 0 < bit_len && bit_len <= (IR_SEND_DATA_MAX_LEN * 8) && code.Length >= code_len_check && repeat_code.Length >= repeat_code_len)
                {
                    // データセット
                    while (true)
                    {
                        outbuffer[0] = 0x00;
                        outbuffer[1] = 0x34;
                        //送信総ビット数
                        outbuffer[2] = (byte)((send_bit_num >> 8) & 0xFF);
                        outbuffer[3] = (byte)(send_bit_num & 0xFF);
                        outbuffer[4] = (byte)((send_bit_pos >> 8) & 0xFF);
                        outbuffer[5] = (byte)(send_bit_pos & 0xFF);
                        if (send_bit_num > send_bit_pos)
                        {
                            set_bit_size = send_bit_num - send_bit_pos;
                            if (set_bit_size > IR_SEND_DATA_USB_SEND_MAX_LEN)
                            {
                                set_bit_size = IR_SEND_DATA_USB_SEND_MAX_LEN;
                            }
                        }
                        else
                        {   // 送信データなし
                            set_bit_size = 0;
                        }
                        outbuffer[6] = (byte)(set_bit_size & 0xFF);

                        if (set_bit_size > 0)
                        {
                            // データセット
                            // 赤外線コードコピー
                            for (uint fi = 0; fi < set_bit_size; fi++)
                            {
                                if (send_bit_pos == 0)
                                {   // Reader Code
                                    // ON Count
                                    outbuffer[7 + (fi * 4)] = (byte)((reader_code >> 24) & 0xFF);
                                    outbuffer[7 + (fi * 4) + 1] = (byte)((reader_code >> 16) & 0xFF);
                                    // OFF Count
                                    outbuffer[7 + (fi * 4) + 2] = (byte)((reader_code >> 8) & 0xFF);
                                    outbuffer[7 + (fi * 4) + 3] = (byte)(reader_code & 0xFF);
                                }
                                else if (send_bit_pos == (send_bit_num - 1 - (repeat_code_len * repeat_code_send_num)))
                                {   // Stop Code
                                    // ON Count
                                    outbuffer[7 + (fi * 4)] = (byte)((stop_code >> 24) & 0xFF);
                                    outbuffer[7 + (fi * 4) + 1] = (byte)((stop_code >> 16) & 0xFF);
                                    // OFF Count
                                    outbuffer[7 + (fi * 4) + 2] = (byte)((stop_code >> 8) & 0xFF);
                                    outbuffer[7 + (fi * 4) + 3] = (byte)(stop_code & 0xFF);
                                }
                                else if (send_bit_pos >= (send_bit_num - (repeat_code_len * repeat_code_send_num)))
                                {   // リピート Code
                                    int tmp_idx = (int)((send_bit_pos - (send_bit_num - (repeat_code_len * repeat_code_send_num))) % repeat_code_len);
                                    // ON Count
                                    outbuffer[7 + (fi * 4)] = (byte)((repeat_code[tmp_idx] >> 24) & 0xFF);
                                    outbuffer[7 + (fi * 4) + 1] = (byte)((repeat_code[tmp_idx] >> 16) & 0xFF);
                                    // OFF Count
                                    outbuffer[7 + (fi * 4) + 2] = (byte)((repeat_code[tmp_idx] >> 8) & 0xFF);
                                    outbuffer[7 + (fi * 4) + 3] = (byte)(repeat_code[tmp_idx] & 0xFF);
                                }
                                else
                                {
                                    int byte_pos = (int)(send_bit_pos - 1) / 8;
                                    int bit_pos = (int)(send_bit_pos - 1) % 8;
                                    int bit_mask = 0x01 << bit_pos;
                                    int tmp_data = code[byte_pos];

                                    if ((tmp_data & bit_mask & 0xFF) != 0)
                                    {   // Bit 1
                                        // ON Count
                                        outbuffer[7 + (fi * 4)] = (byte)((bit_1 >> 24) & 0xFF);
                                        outbuffer[7 + (fi * 4) + 1] = (byte)((bit_1 >> 16) & 0xFF);
                                        // OFF Count
                                        outbuffer[7 + (fi * 4) + 2] = (byte)((bit_1 >> 8) & 0xFF);
                                        outbuffer[7 + (fi * 4) + 3] = (byte)(bit_1 & 0xFF);
                                    }
                                    else
                                    {   // bit 0
                                        // ON Count
                                        outbuffer[7 + (fi * 4)] = (byte)((bit_0 >> 24) & 0xFF);
                                        outbuffer[7 + (fi * 4) + 1] = (byte)((bit_0 >> 16) & 0xFF);
                                        // OFF Count
                                        outbuffer[7 + (fi * 4) + 2] = (byte)((bit_0 >> 8) & 0xFF);
                                        outbuffer[7 + (fi * 4) + 3] = (byte)(bit_0 & 0xFF);
                                    }
                                }
                                send_bit_pos++;
                            }
                            if (WriteFile(HandleToUSBDevice, outbuffer, 65, ref BytesWritten, IntPtr.Zero))
                            {
                                //Now get the response packet from the firmware.
                                inbuffer[0] = 0;
                                {
                                    if (ReadFileManagedBuffer(HandleToUSBDevice, inbuffer, 65, ref BytesRead, IntPtr.Zero))	//Blocking function, unless an "overlapped" structure is used
                                    {
                                        //INBuffer[0] is the report ID, which we don't care about.
                                        //INBuffer[1] is an echo back of the command (see microcontroller firmware).
                                        //INBuffer[2] contains the I/O port pin value for the pushbutton (see microcontroller firmware).  
                                        if (inbuffer[1] == 0x34)
                                        {
                                            if (inbuffer[2] == 0x00)
                                            {   // OK
                                            }
                                            else
                                            {
                                                // NG
                                                error_flag = true;
                                            }
                                        }
                                    }
                                    else
                                    {
                                        // NG
                                        error_flag = true;
                                    }
                                }
                            }
                            else
                            {
                                // NG
                                error_flag = true;
                            }

                        }
                        else
                        {   // 送信データなし
                            break;
                        }
                    }

                    // データ送信要求セット
                    if (error_flag == false)
                    {
                        outbuffer[0] = 0;			//The first byte is the "Report ID" and does not get sent over the USB bus.  Always set = 0.
                        outbuffer[1] = 0x35;		//0x81 is the "Get Pushbutton State" command in the firmware
                        outbuffer[2] = (byte)((freq >> 8) & 0xFF);
                        outbuffer[3] = (byte)(freq & 0xFF);
                        outbuffer[4] = (byte)((send_bit_num >> 8) & 0xFF);
                        outbuffer[5] = (byte)(send_bit_num & 0xFF);

                        //To get the pushbutton state, first, we send a packet with our "Get Pushbutton State" command in it.
                        if (WriteFile(HandleToUSBDevice, outbuffer, 65, ref BytesWritten, IntPtr.Zero))	//Blocking function, unless an "overlapped" structure is used
                        {
                            //Now get the response packet from the firmware.
                            inbuffer[0] = 0;
                            {
                                if (ReadFileManagedBuffer(HandleToUSBDevice, inbuffer, 65, ref BytesRead, IntPtr.Zero))	//Blocking function, unless an "overlapped" structure is used
                                {
                                    //INBuffer[0] is the report ID, which we don't care about.
                                    //INBuffer[1] is an echo back of the command (see microcontroller firmware).
                                    //INBuffer[2] contains the I/O port pin value for the pushbutton (see microcontroller firmware).  
                                    if (inbuffer[1] == 0x35)
                                    {
                                        if (inbuffer[2] == 0x00)
                                        {   // OK
                                            i_ret = 0;
                                        }
                                        else
                                        {
                                            // NG
                                            error_flag = true;
                                        }
                                    }
                                }
                                else
                                {
                                    // NG
                                    i_ret = -5;
                                }
                            }
                        }
                        else
                        {
                            // NG
                            i_ret = -4;
                        }
                    }
                    else
                    {   // データセットエラー
                        i_ret = -3;
                    }
                }
                else
                {   // パラメータエラー
                    i_ret = -2;
                }
            }
            catch
            {
                i_ret = 1;
            }
            return i_ret;
        }
        public static int writeUSBIRData(SafeFileHandle HandleToUSBDevice, uint freq, uint[] data, uint bit_len)
        {
            int i_ret = -1;
            byte[] outbuffer = new byte[65];
            byte[] inbuffer = new byte[65];
            uint BytesWritten = 0;
            uint BytesRead = 0;
            bool error_flag = false;
            uint send_bit_num = bit_len;            // 送信ビット数　リーダーコード + コード + 終了コード
            uint send_bit_pos = 0;                  // 送信セット済みビット位置
            uint set_bit_size = 0;

            try
            {
                // パラメータチェック
                if (HandleToUSBDevice != null && IR_FREQ_MIN <= freq && freq <= IR_FREQ_MAX && 0 < bit_len && bit_len <= (IR_SEND_DATA_MAX_LEN * 8) & data.Length >= (bit_len * 2))
                {
                    // データセット
                    while (true)
                    {
                        outbuffer[0] = 0x00;
                        outbuffer[1] = 0x34;
                        //送信総ビット数
                        outbuffer[2] = (byte)((send_bit_num >> 8) & 0xFF);
                        outbuffer[3] = (byte)(send_bit_num & 0xFF);
                        outbuffer[4] = (byte)((send_bit_pos >> 8) & 0xFF);
                        outbuffer[5] = (byte)(send_bit_pos & 0xFF);
                        if (send_bit_num > send_bit_pos)
                        {
                            set_bit_size = send_bit_num - send_bit_pos;
                            if (set_bit_size > IR_SEND_DATA_USB_SEND_MAX_LEN)
                            {
                                set_bit_size = IR_SEND_DATA_USB_SEND_MAX_LEN;
                            }
                        }
                        else
                        {   // 送信データなし
                            set_bit_size = 0;
                        }
                        outbuffer[6] = (byte)(set_bit_size & 0xFF);

                        if (set_bit_size > 0)
                        {
                            // データセット
                            // 赤外線コードコピー
                            for (uint fi = 0; fi < set_bit_size; fi++)
                            {
                                // ON Count
                                outbuffer[7 + (fi * 4)] = (byte)((data[send_bit_pos * 2] >> 8) & 0xFF);
                                outbuffer[7 + (fi * 4) + 1] = (byte)(data[send_bit_pos * 2] & 0xFF);
                                // OFF Count
                                outbuffer[7 + (fi * 4) + 2] = (byte)((data[(send_bit_pos * 2) + 1] >> 8) & 0xFF);
                                outbuffer[7 + (fi * 4) + 3] = (byte)(data[(send_bit_pos * 2) + 1] & 0xFF);
                                send_bit_pos++;
                            }
                            if (WriteFile(HandleToUSBDevice, outbuffer, 65, ref BytesWritten, IntPtr.Zero))
                            {
                                //Now get the response packet from the firmware.
                                inbuffer[0] = 0;
                                {
                                    if (ReadFileManagedBuffer(HandleToUSBDevice, inbuffer, 65, ref BytesRead, IntPtr.Zero))	//Blocking function, unless an "overlapped" structure is used
                                    {
                                        //INBuffer[0] is the report ID, which we don't care about.
                                        //INBuffer[1] is an echo back of the command (see microcontroller firmware).
                                        //INBuffer[2] contains the I/O port pin value for the pushbutton (see microcontroller firmware).  
                                        if (inbuffer[1] == 0x34)
                                        {
                                            if (inbuffer[2] == 0x00)
                                            {   // OK
                                            }
                                            else
                                            {
                                                // NG
                                                error_flag = true;
                                            }
                                        }
                                    }
                                    else
                                    {
                                        // NG
                                        error_flag = true;
                                    }
                                }
                            }
                            else
                            {
                                // NG
                                error_flag = true;
                            }

                        }
                        else
                        {   // 送信データなし
                            break;
                        }
                    }

                    // データ送信要求セット
                    if (error_flag == false)
                    {
                        outbuffer[0] = 0;			//The first byte is the "Report ID" and does not get sent over the USB bus.  Always set = 0.
                        outbuffer[1] = 0x35;		//0x81 is the "Get Pushbutton State" command in the firmware
                        outbuffer[2] = (byte)((freq >> 8) & 0xFF);
                        outbuffer[3] = (byte)(freq & 0xFF);
                        outbuffer[4] = (byte)((send_bit_num >> 8) & 0xFF);
                        outbuffer[5] = (byte)(send_bit_num & 0xFF);

                        //To get the pushbutton state, first, we send a packet with our "Get Pushbutton State" command in it.
                        if (WriteFile(HandleToUSBDevice, outbuffer, 65, ref BytesWritten, IntPtr.Zero))	//Blocking function, unless an "overlapped" structure is used
                        {
                            //Now get the response packet from the firmware.
                            inbuffer[0] = 0;
                            {
                                if (ReadFileManagedBuffer(HandleToUSBDevice, inbuffer, 65, ref BytesRead, IntPtr.Zero))	//Blocking function, unless an "overlapped" structure is used
                                {
                                    //INBuffer[0] is the report ID, which we don't care about.
                                    //INBuffer[1] is an echo back of the command (see microcontroller firmware).
                                    //INBuffer[2] contains the I/O port pin value for the pushbutton (see microcontroller firmware).  
                                    if (inbuffer[1] == 0x35)
                                    {
                                        if (inbuffer[2] == 0x00)
                                        {   // OK
                                            i_ret = 0;
                                        }
                                        else
                                        {
                                            // NG
                                            error_flag = true;
                                        }
                                    }
                                }
                                else
                                {
                                    // NG
                                    i_ret = -5;
                                }
                            }
                        }
                        else
                        {
                            // NG
                            i_ret = -4;
                        }
                    }
                    else
                    {   // データセットエラー
                        i_ret = -3;
                    }
                }
                else
                {   // パラメータエラー
                    i_ret = -2;
                }
            }
            catch
            {
                i_ret = 1;
            }
            return i_ret;
        }

        public static int writeUSBIRData(SafeFileHandle HandleToUSBDevice, uint freq, byte[] data, uint bit_len)
        {
            int i_ret = -1;
            byte[] outbuffer = new byte[65];
            byte[] inbuffer = new byte[65];
            uint BytesWritten = 0;
            uint BytesRead = 0;
            bool error_flag = false;
            uint send_bit_num = bit_len;            // 送信ビット数　リーダーコード + コード + 終了コード
            uint send_bit_pos = 0;                  // 送信セット済みビット位置
            uint set_bit_size = 0;

            try
            {
                // パラメータチェック
                if (HandleToUSBDevice != null && IR_FREQ_MIN <= freq && freq <= IR_FREQ_MAX && 0 < bit_len && bit_len <= (IR_SEND_DATA_MAX_LEN * 8) & data.Length >= (bit_len * 4))
                {
                    // データセット
                    while (true)
                    {
                        outbuffer[0] = 0x00;
                        outbuffer[1] = 0x34;
                        //送信総ビット数
                        outbuffer[2] = (byte)((send_bit_num >> 8) & 0xFF);
                        outbuffer[3] = (byte)(send_bit_num & 0xFF);
                        outbuffer[4] = (byte)((send_bit_pos >> 8) & 0xFF);
                        outbuffer[5] = (byte)(send_bit_pos & 0xFF);
                        if (send_bit_num > send_bit_pos)
                        {
                            set_bit_size = send_bit_num - send_bit_pos;
                            if (set_bit_size > IR_SEND_DATA_USB_SEND_MAX_LEN)
                            {
                                set_bit_size = IR_SEND_DATA_USB_SEND_MAX_LEN;
                            }
                        }
                        else
                        {   // 送信データなし
                            set_bit_size = 0;
                        }
                        outbuffer[6] = (byte)(set_bit_size & 0xFF);

                        if (set_bit_size > 0)
                        {
                            // データセット
                            // 赤外線コードコピー
                            for (uint fi = 0; fi < set_bit_size; fi++)
                            {
                                // ON Count
                                outbuffer[7 + (fi * 4)] = data[send_bit_pos * 4];
                                outbuffer[7 + (fi * 4) + 1] = data[(send_bit_pos * 4) + 1];
                                // OFF Count
                                outbuffer[7 + (fi * 4) + 2] = data[(send_bit_pos * 4) + 2];
                                outbuffer[7 + (fi * 4) + 3] = data[(send_bit_pos * 4) + 3];
                                send_bit_pos++;
                            }
                            if (WriteFile(HandleToUSBDevice, outbuffer, 65, ref BytesWritten, IntPtr.Zero))
                            {
                                //Now get the response packet from the firmware.
                                inbuffer[0] = 0;
                                {
                                    if (ReadFileManagedBuffer(HandleToUSBDevice, inbuffer, 65, ref BytesRead, IntPtr.Zero))	//Blocking function, unless an "overlapped" structure is used
                                    {
                                        //INBuffer[0] is the report ID, which we don't care about.
                                        //INBuffer[1] is an echo back of the command (see microcontroller firmware).
                                        //INBuffer[2] contains the I/O port pin value for the pushbutton (see microcontroller firmware).  
                                        if (inbuffer[1] == 0x34)
                                        {
                                            if (inbuffer[2] == 0x00)
                                            {   // OK
                                            }
                                            else
                                            {
                                                // NG
                                                error_flag = true;
                                            }
                                        }
                                    }
                                    else
                                    {
                                        // NG
                                        error_flag = true;
                                    }
                                }
                            }
                            else
                            {
                                // NG
                                error_flag = true;
                            }

                        }
                        else
                        {   // 送信データなし
                            break;
                        }
                    }

                    // データ送信要求セット
                    if (error_flag == false)
                    {
                        outbuffer[0] = 0;			//The first byte is the "Report ID" and does not get sent over the USB bus.  Always set = 0.
                        outbuffer[1] = 0x35;		//0x81 is the "Get Pushbutton State" command in the firmware
                        outbuffer[2] = (byte)((freq >> 8) & 0xFF);
                        outbuffer[3] = (byte)(freq & 0xFF);
                        outbuffer[4] = (byte)((send_bit_num >> 8) & 0xFF);
                        outbuffer[5] = (byte)(send_bit_num & 0xFF);

                        //To get the pushbutton state, first, we send a packet with our "Get Pushbutton State" command in it.
                        if (WriteFile(HandleToUSBDevice, outbuffer, 65, ref BytesWritten, IntPtr.Zero))	//Blocking function, unless an "overlapped" structure is used
                        {
                            //Now get the response packet from the firmware.
                            inbuffer[0] = 0;
                            {
                                if (ReadFileManagedBuffer(HandleToUSBDevice, inbuffer, 65, ref BytesRead, IntPtr.Zero))	//Blocking function, unless an "overlapped" structure is used
                                {
                                    //INBuffer[0] is the report ID, which we don't care about.
                                    //INBuffer[1] is an echo back of the command (see microcontroller firmware).
                                    //INBuffer[2] contains the I/O port pin value for the pushbutton (see microcontroller firmware).  
                                    if (inbuffer[1] == 0x35)
                                    {
                                        if (inbuffer[2] == 0x00)
                                        {   // OK
                                            i_ret = 0;
                                        }
                                        else
                                        {
                                            // NG
                                            error_flag = true;
                                        }
                                    }
                                }
                                else
                                {
                                    // NG
                                    i_ret = -5;
                                }
                            }
                        }
                        else
                        {
                            // NG
                            i_ret = -4;
                        }
                    }
                    else
                    {   // データセットエラー
                        i_ret = -3;
                    }
                }
                else
                {   // パラメータエラー
                    i_ret = -2;
                }
            }
            catch
            {
                i_ret = 1;
            }
            return i_ret;
        }
        public static int writeUSBIR_Plarail_Stop(SafeFileHandle HandleToUSBDevice, PLARAIL_BAND band)
        {
            int i_ret = -1;
            bool b_err = false;
            uint[] code = null;    // 送信データ
            uint[] code_A_STOP = new uint[] { 0x0016, 0x001D, 0x0016, 0x001D, 0x0016, 0x001D, 0x0016, 0x001D, 0x0034, 0x0013, 0x0034, 0x0013, 0x0034, 0x0013, 0x0034, 0x0013, 0x0000, 0x0280,
                                                0x0016, 0x001D, 0x0016, 0x001D, 0x0016, 0x001D, 0x0016, 0x001D, 0x0034, 0x0013, 0x0034, 0x0013, 0x0034, 0x0013, 0x0034, 0x0013, 0x0000, 0x0280};
            uint[] code_B_STOP = new uint[] { 0x0016, 0x001D, 0x0034, 0x0013, 0x0016, 0x001D, 0x0016, 0x001D, 0x0034, 0x0013, 0x0016, 0x001D, 0x0034, 0x0013, 0x0034, 0x0013, 0x0000, 0x0280,
                                                0x0016, 0x001D, 0x0034, 0x0013, 0x0016, 0x001D, 0x0016, 0x001D, 0x0034, 0x0013, 0x0016, 0x001D, 0x0034, 0x0013, 0x0034, 0x0013, 0x0000, 0x0280};
            try
            {
                if (band == PLARAIL_BAND.BAND_A)
                {
                    code = code_A_STOP;
                }
                else if (band == PLARAIL_BAND.BAND_B)
                {
                    code = code_B_STOP;
                }
                else
                {
                    i_ret = 2;
                    b_err = true;
                }

                if (b_err == false)
                {
                    // USB DEVICEへ送信 パラメータ[USB DEVICEハンドル、周波数、送信赤外線コード、赤外線コードのビット長]
                    i_ret = writeUSBIRData(HandleToUSBDevice, 38000, code, 18);
                    if (i_ret != 0)
                    {
                        i_ret = 3;
                    }
                }
            }
            catch
            {
                i_ret = 1;
            }
            return i_ret;
        }
        public static int writeUSBIR_Plarail_Speed_Up(SafeFileHandle HandleToUSBDevice, PLARAIL_BAND band, PLARAIL_DIRECTION dir)
        {
            int i_ret = -1;
            bool b_err = false;
            // 送信データ ON時間とOFF時間の組み合わせ
            uint[] code = null;    // 送信データ
            uint[] code_A_UP_F = new uint[]     { 0x0016, 0x001D, 0x0016, 0x001D, 0x0016, 0x001D, 0x0034, 0x0013, 0x0034, 0x0013, 0x0034, 0x0013, 0x0034, 0x0013, 0x0016, 0x001D, 0x0000, 0x0280,
                                                    0x0016, 0x001D, 0x0016, 0x001D, 0x0016, 0x001D, 0x0034, 0x0013, 0x0034, 0x0013, 0x0034, 0x0013, 0x0034, 0x0013, 0x0016, 0x001D, 0x0000, 0x0280};
            uint[] code_A_UP_B = new uint[]     { 0x0016, 0x001D, 0x0016, 0x001D, 0x0034, 0x0013, 0x0016, 0x001D, 0x0034, 0x0013, 0x0034, 0x0013, 0x0016, 0x001D, 0x0034, 0x0013, 0x0000, 0x0280,
                                                    0x0016, 0x001D, 0x0016, 0x001D, 0x0034, 0x0013, 0x0016, 0x001D, 0x0034, 0x0013, 0x0034, 0x0013, 0x0016, 0x001D, 0x0034, 0x0013, 0x0000, 0x0280};
            uint[] code_B_UP_F = new uint[]     { 0x0016, 0x001D, 0x0034, 0x0013, 0x0016, 0x001D, 0x0034, 0x0013, 0x0034, 0x0013, 0x0016, 0x001D, 0x0034, 0x0013, 0x0016, 0x001D, 0x0000, 0x0280,
                                                    0x0016, 0x001D, 0x0034, 0x0013, 0x0016, 0x001D, 0x0034, 0x0013, 0x0034, 0x0013, 0x0016, 0x001D, 0x0034, 0x0013, 0x0016, 0x001D, 0x0000, 0x0280};
            uint[] code_B_UP_B = new uint[]     { 0x0016, 0x001D, 0x0034, 0x0013, 0x0034, 0x0013, 0x0016, 0x001D, 0x0034, 0x0013, 0x0016, 0x001D, 0x0016, 0x001D, 0x0034, 0x0013, 0x0000, 0x0280,
                                                    0x0016, 0x001D, 0x0034, 0x0013, 0x0034, 0x0013, 0x0016, 0x001D, 0x0034, 0x0013, 0x0016, 0x001D, 0x0016, 0x001D, 0x0034, 0x0013, 0x0000, 0x0280};

            try
            {
                if (band == PLARAIL_BAND.BAND_A)
                {
                    if (dir == PLARAIL_DIRECTION.FORWARD)
                    {
                        code = code_A_UP_F;
                    }
                    else if (dir == PLARAIL_DIRECTION.BACKWARD)
                    {
                        code = code_A_UP_B;
                    }
                    else
                    {
                        i_ret = 3;
                        b_err = true;
                    }
                }
                else if (band == PLARAIL_BAND.BAND_B)
                {
                    if (dir == PLARAIL_DIRECTION.FORWARD)
                    {
                        code = code_B_UP_F;
                    }
                    else if (dir == PLARAIL_DIRECTION.BACKWARD)
                    {
                        code = code_B_UP_B;
                    }
                    else
                    {
                        i_ret = 3;
                        b_err = true;
                    }
                }
                else
                {
                    i_ret = 2;
                    b_err = true;
                }

                if(b_err == false)
                {
                    // USB DEVICEへ送信 パラメータ[USB DEVICEハンドル、周波数、送信赤外線コード、赤外線コードのビット長]
                    i_ret = writeUSBIRData(HandleToUSBDevice, 38000, code, 18);
                    if(i_ret != 0)
                    {
                        i_ret = 4;
                    }
                }
            }
            catch
            {
                i_ret = 1;
            }
            return i_ret;
        }
        public static int writeUSBIR_Plarail_Speed_Down(SafeFileHandle HandleToUSBDevice, PLARAIL_BAND band)
        {
            int i_ret = -1;
            bool b_err = false;
            // 送信データ ON時間とOFF時間の組み合わせ
            uint[] code = null;    // 送信データ
            uint[] code_A_DOWN_F = new uint[] { 0x0016, 0x001D, 0x0016, 0x001D, 0x0034, 0x0013, 0x0034, 0x0013, 0x0034, 0x0013, 0x0034, 0x0013, 0x0016, 0x001D, 0x0016, 0x001D, 0x0000, 0x0280,
                                                0x0016, 0x001D, 0x0016, 0x001D, 0x0034, 0x0013, 0x0034, 0x0013, 0x0034, 0x0013, 0x0034, 0x0013, 0x0016, 0x001D, 0x0016, 0x001D, 0x0000, 0x0280};
            //uint[] code_A_DOWN_B = new uint[] { 0x0016, 0x001D, 0x0016, 0x001D, 0x0034, 0x0013, 0x0034, 0x0013, 0x0034, 0x0013, 0x0034, 0x0013, 0x0016, 0x001D, 0x0016, 0x001D, 0x0000, 0x0280 };
            uint[] code_B_DOWN_F = new uint[] { 0x0016, 0x001D, 0x0034, 0x0013, 0x0034, 0x0013, 0x0034, 0x0013, 0x0034, 0x0013, 0x0016, 0x001D, 0x0016, 0x001D, 0x0016, 0x001D, 0x0000, 0x0280,
                                                0x0016, 0x001D, 0x0034, 0x0013, 0x0034, 0x0013, 0x0034, 0x0013, 0x0034, 0x0013, 0x0016, 0x001D, 0x0016, 0x001D, 0x0016, 0x001D, 0x0000, 0x0280};
            //uint[] code_B_DOWN_B = new uint[] { 0x0016, 0x001D, 0x0034, 0x0013, 0x0034, 0x0013, 0x0034, 0x0013, 0x0034, 0x0013, 0x0016, 0x001D, 0x0016, 0x001D, 0x0016, 0x001D, 0x0000, 0x0280 };

            try
            {
                if (band == PLARAIL_BAND.BAND_A)
                {
                    code = code_A_DOWN_F;
                    //if (dir == PLARAIL_DIRECTION.FORWARD)
                    //{
                    //    code = code_A_DOWN_F;
                    //}
                    //else if (dir == PLARAIL_DIRECTION.BACKWARD)
                    //{
                    //    code = code_B_DOWN_F;
                    //}
                    //else
                    //{
                    //    i_ret = 3;
                    //    b_err = true;
                    //}
                }
                else if (band == PLARAIL_BAND.BAND_B)
                {
                    code = code_B_DOWN_F;
                    //if (dir == PLARAIL_DIRECTION.FORWARD)
                    //{
                    //    code = code_B_DOWN_F;
                    //}
                    //else if (dir == PLARAIL_DIRECTION.BACKWARD)
                    //{
                    //    code = code_B_DOWN_B;
                    //}
                    //else
                    //{
                    //    i_ret = 3;
                    //    b_err = true;
                    //}
                }
                else
                {
                    i_ret = 2;
                    b_err = true;
                }

                if (b_err == false)
                {
                    // USB DEVICEへ送信 パラメータ[USB DEVICEハンドル、周波数、送信赤外線コード、赤外線コードのビット長]
                    i_ret = writeUSBIRData(HandleToUSBDevice, 38000, code, 18);
                    if (i_ret != 0)
                    {
                        i_ret = 3;
                    }
                }
            }
            catch
            {
                i_ret = 1;
            }
            return i_ret;
        }

        //FUNCTION:	CheckIfPresentAndGetUSBDevicePath()
        //PURPOSE:	Check if a USB device is currently plugged in with a matching VID and PID
        //INPUT:	Uses globally declared String DevicePath, globally declared GUID, and the MY_DEVICE_ID constant.
        //OUTPUT:	Returns BOOL.  TRUE when device with matching VID/PID found.  FALSE if device with VID/PID could not be found.
        //			When returns TRUE, the globally accessable "DetailedInterfaceDataStructure" will contain the device path
        //			to the USB device with the matching VID/PID.
        static bool CheckIfPresentAndGetUSBDevicePath()
        {
            /* 
            Before we can "connect" our application to our USB embedded device, we must first find the device.
            A USB bus can have many devices simultaneously connected, so somehow we have to find our device only.
            This is done with the Vendor ID (VID) and Product ID (PID).  Each USB product line should have
            a unique combination of VID and PID.  

            Microsoft has created a number of functions which are useful for finding plug and play devices.  Documentation
            for each function used can be found in the MSDN library.  We will be using the following functions (unmanaged C functions):

            SetupDiGetClassDevs()					//provided by setupapi.dll, which comes with Windows
            SetupDiEnumDeviceInterfaces()			//provided by setupapi.dll, which comes with Windows
            GetLastError()							//provided by kernel32.dll, which comes with Windows
            SetupDiDestroyDeviceInfoList()			//provided by setupapi.dll, which comes with Windows
            SetupDiGetDeviceInterfaceDetail()		//provided by setupapi.dll, which comes with Windows
            SetupDiGetDeviceRegistryProperty()		//provided by setupapi.dll, which comes with Windows
            CreateFile()							//provided by kernel32.dll, which comes with Windows

            In order to call these unmanaged functions, the Marshal class is very useful.
             
            We will also be using the following unusual data types and structures.  Documentation can also be found in
            the MSDN library:

            PSP_DEVICE_INTERFACE_DATA
            PSP_DEVICE_INTERFACE_DETAIL_DATA
            SP_DEVINFO_DATA
            HDEVINFO
            HANDLE
            GUID

            The ultimate objective of the following code is to get the device path, which will be used elsewhere for getting
            read and write handles to the USB device.  Once the read/write handles are opened, only then can this
            PC application begin reading/writing to the USB device using the WriteFile() and ReadFile() functions.

            Getting the device path is a multi-step round about process, which requires calling several of the
            SetupDixxx() functions provided by setupapi.dll.
            */

            try
            {
                IntPtr DeviceInfoTable = IntPtr.Zero;
                SP_DEVICE_INTERFACE_DATA InterfaceDataStructure = new SP_DEVICE_INTERFACE_DATA();
                SP_DEVICE_INTERFACE_DETAIL_DATA DetailedInterfaceDataStructure = new SP_DEVICE_INTERFACE_DETAIL_DATA();
                SP_DEVINFO_DATA DevInfoData = new SP_DEVINFO_DATA();

                uint InterfaceIndex = 0;
                uint dwRegType = 0;
                uint dwRegSize = 0;
                uint dwRegSize2 = 0;
                uint StructureSize = 0;
                IntPtr PropertyValueBuffer = IntPtr.Zero;
                bool MatchFound = false;
                uint ErrorStatus;
                uint LoopCounter = 0;

                //Use the formatting: "Vid_xxxx&Pid_xxxx" where xxxx is a 16-bit hexadecimal number.
                //Make sure the value appearing in the parathesis matches the USB device descriptor
                //of the device that this aplication is intending to find.
                String DeviceIDToFind = "Vid_22ea&Pid_003A";
#if COMPOSITE_DEVICE
                bool MatchFound2 = false;
                String DeviceIDToFind2 = "Mi_03";
#else
#endif

                //First populate a list of plugged in devices (by specifying "DIGCF_PRESENT"), which are of the specified class GUID. 
                DeviceInfoTable = SetupDiGetClassDevs(ref InterfaceClassGuid, IntPtr.Zero, IntPtr.Zero, DIGCF_PRESENT | DIGCF_DEVICEINTERFACE);

                if (DeviceInfoTable != IntPtr.Zero)
                {
                    //Now look through the list we just populated.  We are trying to see if any of them match our device. 
                    while (true)
                    {
                        InterfaceDataStructure.cbSize = (uint)Marshal.SizeOf(InterfaceDataStructure);
                        if (SetupDiEnumDeviceInterfaces(DeviceInfoTable, IntPtr.Zero, ref InterfaceClassGuid, InterfaceIndex, ref InterfaceDataStructure))
                        {
                            ErrorStatus = (uint)Marshal.GetLastWin32Error();
                            if (ErrorStatus == ERROR_NO_MORE_ITEMS)	//Did we reach the end of the list of matching devices in the DeviceInfoTable?
                            {	//Cound not find the device.  Must not have been attached.
                                SetupDiDestroyDeviceInfoList(DeviceInfoTable);	//Clean up the old structure we no longer need.
                                return false;
                            }
                        }
                        else	//Else some other kind of unknown error ocurred...
                        {
                            ErrorStatus = (uint)Marshal.GetLastWin32Error();
                            SetupDiDestroyDeviceInfoList(DeviceInfoTable);	//Clean up the old structure we no longer need.
                            return false;
                        }

                        //Now retrieve the hardware ID from the registry.  The hardware ID contains the VID and PID, which we will then 
                        //check to see if it is the correct device or not.

                        //Initialize an appropriate SP_DEVINFO_DATA structure.  We need this structure for SetupDiGetDeviceRegistryProperty().
                        DevInfoData.cbSize = (uint)Marshal.SizeOf(DevInfoData);
                        SetupDiEnumDeviceInfo(DeviceInfoTable, InterfaceIndex, ref DevInfoData);

                        //First query for the size of the hardware ID, so we can know how big a buffer to allocate for the data.
                        SetupDiGetDeviceRegistryProperty(DeviceInfoTable, ref DevInfoData, SPDRP_HARDWAREID, ref dwRegType, IntPtr.Zero, 0, ref dwRegSize);

                        //Allocate a buffer for the hardware ID.
                        //Should normally work, but could throw exception "OutOfMemoryException" if not enough resources available.
                        PropertyValueBuffer = Marshal.AllocHGlobal((int)dwRegSize);

                        //Retrieve the hardware IDs for the current device we are looking at.  PropertyValueBuffer gets filled with a 
                        //REG_MULTI_SZ (array of null terminated strings).  To find a device, we only care about the very first string in the
                        //buffer, which will be the "device ID".  The device ID is a string which contains the VID and PID, in the example 
                        //format "Vid_04d8&Pid_003f".
                        SetupDiGetDeviceRegistryProperty(DeviceInfoTable, ref DevInfoData, SPDRP_HARDWAREID, ref dwRegType, PropertyValueBuffer, dwRegSize, ref dwRegSize2);

                        //Now check if the first string in the hardware ID matches the device ID of the USB device we are trying to find.
                        String DeviceIDFromRegistry = Marshal.PtrToStringUni(PropertyValueBuffer); //Make a new string, fill it with the contents from the PropertyValueBuffer

                        Marshal.FreeHGlobal(PropertyValueBuffer);		//No longer need the PropertyValueBuffer, free the memory to prevent potential memory leaks

                        //Convert both strings to lower case.  This makes the code more robust/portable accross OS Versions
                        DeviceIDFromRegistry = DeviceIDFromRegistry.ToLowerInvariant();
#if COMPOSITE_DEVICE
                        DeviceIDToFind = DeviceIDToFind.ToLowerInvariant();
                        DeviceIDToFind2 = DeviceIDToFind2.ToLowerInvariant();

                        //Now check if the hardware ID we are looking at contains the correct VID/PID
                        MatchFound = DeviceIDFromRegistry.Contains(DeviceIDToFind);
                        MatchFound2 = DeviceIDFromRegistry.Contains(DeviceIDToFind2);

                        if ((MatchFound == true) && (MatchFound2 == true))
#else
                        DeviceIDToFind = DeviceIDToFind.ToLowerInvariant();

                        //Now check if the hardware ID we are looking at contains the correct VID/PID
                        MatchFound = DeviceIDFromRegistry.Contains(DeviceIDToFind);

                        if (MatchFound == true)
#endif
                        {
                            //Device must have been found.  In order to open I/O file handle(s), we will need the actual device path first.
                            //We can get the path by calling SetupDiGetDeviceInterfaceDetail(), however, we have to call this function twice:  The first
                            //time to get the size of the required structure/buffer to hold the detailed interface data, then a second time to actually 
                            //get the structure (after we have allocated enough memory for the structure.)
                            DetailedInterfaceDataStructure.cbSize = (uint)Marshal.SizeOf(DetailedInterfaceDataStructure);
                            //First call populates "StructureSize" with the correct value
                            SetupDiGetDeviceInterfaceDetail(DeviceInfoTable, ref InterfaceDataStructure, IntPtr.Zero, 0, ref StructureSize, IntPtr.Zero);
                            //Need to call SetupDiGetDeviceInterfaceDetail() again, this time specifying a pointer to a SP_DEVICE_INTERFACE_DETAIL_DATA buffer with the correct size of RAM allocated.
                            //First need to allocate the unmanaged buffer and get a pointer to it.
                            IntPtr pUnmanagedDetailedInterfaceDataStructure = IntPtr.Zero;  //Declare a pointer.
                            pUnmanagedDetailedInterfaceDataStructure = Marshal.AllocHGlobal((int)StructureSize);    //Reserve some unmanaged memory for the structure.
                            DetailedInterfaceDataStructure.cbSize = 6; //Initialize the cbSize parameter (4 bytes for DWORD + 2 bytes for unicode null terminator)
                            Marshal.StructureToPtr(DetailedInterfaceDataStructure, pUnmanagedDetailedInterfaceDataStructure, false); //Copy managed structure contents into the unmanaged memory buffer.

                            //Now call SetupDiGetDeviceInterfaceDetail() a second time to receive the device path in the structure at pUnmanagedDetailedInterfaceDataStructure.
                            if (SetupDiGetDeviceInterfaceDetail(DeviceInfoTable, ref InterfaceDataStructure, pUnmanagedDetailedInterfaceDataStructure, StructureSize, IntPtr.Zero, IntPtr.Zero))
                            {
                                //Need to extract the path information from the unmanaged "structure".  The path starts at (pUnmanagedDetailedInterfaceDataStructure + sizeof(DWORD)).
                                IntPtr pToDevicePath = new IntPtr((uint)pUnmanagedDetailedInterfaceDataStructure.ToInt32() + 4);  //Add 4 to the pointer (to get the pointer to point to the path, instead of the DWORD cbSize parameter)
                                DevicePath = Marshal.PtrToStringUni(pToDevicePath); //Now copy the path information into the globally defined DevicePath String.

                                //We now have the proper device path, and we can finally use the path to open I/O handle(s) to the device.
                                SetupDiDestroyDeviceInfoList(DeviceInfoTable);	//Clean up the old structure we no longer need.
                                Marshal.FreeHGlobal(pUnmanagedDetailedInterfaceDataStructure);  //No longer need this unmanaged SP_DEVICE_INTERFACE_DETAIL_DATA buffer.  We already extracted the path information.
                                return true;    //Returning the device path in the global DevicePath String
                            }
                            else //Some unknown failure occurred
                            {
                                uint ErrorCode = (uint)Marshal.GetLastWin32Error();
                                SetupDiDestroyDeviceInfoList(DeviceInfoTable);	//Clean up the old structure.
                                Marshal.FreeHGlobal(pUnmanagedDetailedInterfaceDataStructure);  //No longer need this unmanaged SP_DEVICE_INTERFACE_DETAIL_DATA buffer.  We already extracted the path information.
                                return false;
                            }
                        }

                        InterfaceIndex++;
                        //Keep looping until we either find a device with matching VID and PID, or until we run out of devices to check.
                        //However, just in case some unexpected error occurs, keep track of the number of loops executed.
                        //If the number of loops exceeds a very large number, exit anyway, to prevent inadvertent infinite looping.
                        LoopCounter++;
                        if (LoopCounter == 10000000)	//Surely there aren't more than 10 million devices attached to any forseeable PC...
                        {
                            return false;
                        }
                    }//end of while(true)
                }
                return false;
            }//end of try
            catch
            {
                //Something went wrong if PC gets here.  Maybe a Marshal.AllocHGlobal() failed due to insufficient resources or something.
                return false;
            }
        }


        //--------------------------------------------------------------------------------------------------------------------------
        //FUNCTION:	ReadFileManagedBuffer()
        //PURPOSE:	Wrapper function to call ReadFile()
        //
        //INPUT:	Uses managed versions of the same input parameters as ReadFile() uses.
        //
        //OUTPUT:	Returns boolean indicating if the function call was successful or not.
        //          Also returns data in the byte[] INBuffer, and the number of bytes read. 
        //
        //Notes:    Wrapper function used to call the ReadFile() function.  ReadFile() takes a pointer to an unmanaged buffer and deposits
        //          the bytes read into the buffer.  However, can't pass a pointer to a managed buffer directly to ReadFile().
        //          This ReadFileManagedBuffer() is a wrapper function to make it so application code can call ReadFile() easier
        //          by specifying a managed buffer.
        //--------------------------------------------------------------------------------------------------------------------------
        static public unsafe bool ReadFileManagedBuffer(SafeFileHandle hFile, byte[] INBuffer, uint nNumberOfBytesToRead, ref uint lpNumberOfBytesRead, IntPtr lpOverlapped)
        {
            IntPtr pINBuffer = IntPtr.Zero;

            try
            {
                pINBuffer = Marshal.AllocHGlobal((int)nNumberOfBytesToRead);    //Allocate some unmanged RAM for the receive data buffer.

                if (ReadFile(hFile, pINBuffer, nNumberOfBytesToRead, ref lpNumberOfBytesRead, lpOverlapped))
                {
                    Marshal.Copy(pINBuffer, INBuffer, 0, (int)lpNumberOfBytesRead);    //Copy over the data from unmanged memory into the managed byte[] INBuffer
                    Marshal.FreeHGlobal(pINBuffer);
                    return true;
                }
                else
                {
                    Marshal.FreeHGlobal(pINBuffer);
                    return false;
                }

            }
            catch
            {
                if (pINBuffer != IntPtr.Zero)
                {
                    Marshal.FreeHGlobal(pINBuffer);
                }
                return false;
            }
        }
    }
}
